#!/usr/bin/perl 
###############################################################################
#
# Make a makefile for synthesis
#
# Use -help for usage
#
# Originally written by Costas Calamvokis.  
# Subsequently hacked by Paul Zimmer (paulzimmer@zimmerdesignservices.com).
#
###############################################################################

# check the perl version is high enough
if ( $] < 5.004 ) {
    print "Error: this script can only run with Perl version 5.004 or greater\n";
    exit 1;
}

# Make sure calling location is in INC path
use File::Basename;
$my_dir = &dirname($0);
$me = &basename($0);
unshift(@INC,"$my_dir");

# The "or die" here doesn't work - require dies first...
require rvp or die "This script requires rvp - the rough verilog parser.  It is available at http://www.burbleland.com/v2html/rvp.html";

use vars qw($debug @files @inc_dirs @lib_dirs $lib_ext $vdb @hier_tops
            @db_files $mod $f %global_done );

# defaults
$debug=0; 
@inc_dirs=('.');
@lib_dirs=();
$lib_ext='';

# process options
&process_options;
print "# This makefile was automatically generated by $0\n";

print "\n";

# pass in: list of files on command line
#          empty list of global includes []
#          empty hash of predefined defines {}
#          and set quiet to be quiet (1)
#          then incdirs, libdirs, libext
$vdb = &rvp::read_verilog(\@files,[],{},1,\@inc_dirs,\@lib_dirs,$lib_ext);
my @problems = &rvp::get_problems($vdb);
if (@problems) {
    foreach my $problem (&rvp::get_problems($vdb)) {
	print STDERR "$problem.\n";
    }
    # die "Warnings parsing files!";
}


print "#  Source files scanned:\n";
print "#  ".join("\n#  ",&rvp::get_files($vdb)) . "\n\n";

# find modules at the top of the hierarchy
foreach $mod (&rvp::get_modules($vdb)) {
    print "  checking if module $mod is a top\n" if $debug;
    if  (! &rvp::get_first_instantiator($vdb,$mod)) {
	print "      $mod it is!\n" if $debug;
	push(@hier_tops,$mod);
    }
}

# first target is all modules at top of hierarchy
print "all:";
foreach $mod (@hier_tops) {    print " ${mod}.db"; }
print "\n\n";

# now do all the dependencies working down the hierarchy
foreach $mod (@hier_tops) { 
    &do_deps($vdb,$mod);
}

# Print a list of DB files and the clean rule
print "DB_FILES=".join(" ",@db_files)."\n\n";

print "clean:\n";
print "\trm -f ".'${DB_FILES} ${DB_FILES:db=synlog}'."\n\n";

# Try to solve the annoying gmake/RCS problem
print "
# These lines attempt to force gmake to not mess about with RCS...
%.${scr}: ;
%.v: ;

";

exit 0;

#################################################################################
#
# Subroutines

sub do_deps {
    my ($vdb,$mod) = @_;
    my ($f,@clocks,$sig,$s_line,$s_a_line,$s_i_line,$s_type,$s_file,
	$inc,%done,@instances,$imod,$mf,$iname,$l,$ff,
	$file_list,$clks,@submods,%inc_done,@traverse,$mm);

    ($f) = &rvp::get_modules_file($vdb,$mod);
    die "Couldn't find file for module $mod" unless $f;

    # print the target name
    print "${mod}.db: \\\n";

    # print the dependencies
    #   print the source file's full name
    print "\t" . &rvp::get_files_full_name($vdb,$f) . " \\\n";
    #   print a list of all the included files
    foreach $inc (&rvp::get_files_includes($vdb,$f)) {
	if (!exists($inc_done{$inc})) {
	    print "\t" . &rvp::get_files_full_name($vdb,$inc) . " \\\n";
	    $inc_done{$inc}=1;
	}
    }

    # add this db file to the list for cleaning
    push(@db_files,"${mod}.db");
    
    undef(%done);
    @instances = ();
    @submods = ();

    # do a list of .db files it depends on
    #  go through all of the things this modules instantiates
    for (($imod,$mf,$iname,$l) = &rvp::get_first_instantiation($vdb,$mod );
	 $imod;
	 ($imod,$mf,$iname,$l) = &rvp::get_next_instantiation($vdb)) {
	push(@instances,$iname);
	# Only print the dependency once for each submodule
	if (!exists($done{$imod})) {
	    $done{$imod} = $iname;
	    print "\t$imod.db \\\n";
	}
	# only push this onto the submods list if it
	#  hasn't already been done before (anywhere)
	if (!exists($global_done{$imod})) {
	    $global_done{$imod}=1;
	    push(@submods,$imod);
	}
    }

    # Add any default dependencies
    foreach $default_dep (@default_deps) {
      eval("print \"\t$default_dep \\\\\n\"");
    }

    print "\n";
    # now print out the command to run.
    eval("print \"\t$dcsh\n\"");
    print "\n";

    # now do all the submodules
    foreach $imod (@submods) {
	&do_deps($vdb,$imod);
    }

}

sub process_options {
  use Getopt::Long;
  # Default non-null options
  $scr = "scr";
  $synlog = "synlog";
  $dcsh = "dc_shell -f \${mod}.\${scr} > \${mod}.\${synlog}";
  die unless GetOptions(
    "help" => \$help,
    "verbose" => \$verbose,
    "debug" => \$debug,
    "incdir=s" => \@inc_dirs,
    "libext=s" => \$lib_ext,
    "scr=s" => \$scr,
    "dcsh=s" => \$dcsh,
    "synlog=s" => \$synlog,
    "default_deps=s" => \@default_deps,
    );
  
  @files = @ARGV;
  
  if ($help || !(@files)) {
    &show_help;
    exit;
  }
}


sub show_help {
 print STDERR "Syntax:
  $me [options] *.v

  options:
  -help                 Show this help
  -verbose              Be Chatty
  -debug                Turn on debug messages
  -incdir \"incdir\"    Add \"incdir\" to the list of include directories
                        (can be used multiple times)
  -libext \"libext\"    Add \"libext\" to the list of library extensions
                        recognized by the parser.
                        (can be used multiple times)
  -scr \"suffix\"       Use \"suffix\" as script suffix
                        (defaults to \"$scr\")
  -synlog \"synlog\"    Use \"synlog\" as synthesis log file suffix
                        (defaults to \"$synlog\")
  -dcsh \"command\"     Use \"command\" as the command to build a module.
                        This will be eval'ed by perl.  Var names available
                        include:

                        \${mod}            module being built
                        \${scr}          script suffix (from -scr or default)

                        defaults to \"$dcsh\"
  -default_deps \"dep\"  Use \"dep\" as a default dependency for all modules.
                        This would be used for the .synopsys_dc.setup file, as
                        well as any other script files shared by all modules.
                        This will be eval'ed by perl.  Var names available
                        are the same as for -dcsh option.

  Example:
  $me \\
  -scr dctl \\
  -synlog synlog \\
  -dcsh \'\\\@echo Compiling design : \${mod}\\n\\ttouch \${mod}.start ; dc_shell -f \${mod} > \${mod}.${synlog} ; touch \${mod}.done' \\
  -default_dep \'.synopsys_dc.setup\' \\
  -default_dep \'environment.dctl\' \\
  -default_dep \'common.dctl\' \\
  -default_dep \'common.const.dctl\' \\
  -default_dep \'do_\${mod}\' \\
  *.v

  NOTE: This script requires rvp - the rough verilog parser.  It is available at http://www.burbleland.com/v2html/rvp.html

  \n";
}

